; vim: sts=4 sw=8 et ai fdm=marker

;
;                 fnordlicht firmware
;
;    for additional information please
;    see http://koeln.ccc.de/prozesse/running/fnordlicht
;
; (c) by Alexander Neumann <alexander@bumpern.de>
;    and Lars Noschinski <lars@public.noschinski.de>
;
; This program is free software; you can redistribute it and/or modify
; it under the terms of the GNU General Public License version 2 as
; published by the Free Software Foundation.
;
; This program is distributed in the hope that it will be useful,
; but WITHOUT ANY WARRANTY; without even the implied warranty of
; MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
; GNU General Public License for more details.
;
; You should have received a copy of the GNU General Public License
; along with this program; if not, write to the Free Software
; Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
;
; For more information on the GPL, please go to:
; http://www.gnu.org/copyleft/gpl.html



.nolist
.include "m8def.inc"
.list
.include "fifo_handling.inc"
.include "stack_handling.inc"

;   compile settings
; =============================================================================
; .equ SERIALDEBUG = 1               ; Set this for serial debugging output



;   eeprom settings, upload with 'make install-eeprom'
; =============================================================================
.equ I2CADDR            =   0b0000100 ; i2c addresses, 7bit
.equ I2CGCE             =   1       ; general call enabled

.eseg
        EEPROM_I2CADDR:     .db (I2CADDR<<1)|(I2CGCE)
.dseg



;   flags and aliases {{{
; =============================================================================
; flags (in register Flags)
.equ    PWMPulse0       = 0         ; if bit0 is set, it's the first pwm pulse
.equ    PWMLastPulse    = 1         ; if bit1 is set, we're in the last pwm pulse
                                    ; set by:   timer1 interrupt
                                    ; reset by: main loop
.equ    PWMPulseRelax   = 3         ; if this flag is set, pwm interrupts are
                                    ; not so time critical anymore
                                    ; set by:   timer1 interrupt
                                    ; reset by: main loop
.equ    ScriptEnabled   = 2         ; if bit2 is set, scripts are enabled

; PWM properties
.equ    PWMChannels     =   3
.equ    PWMLevels       =   255     ; max brightness value
.equ    PWMBurstLevels  =   16      ; burst pulses count

; PWM output channel bits
.equ    PWMChannel0     =   0
.equ    PWMChannel1     =   1
.equ    PWMChannel2     =   2

; LedStatusTable offsets
.equ    LedRecBrightness = 0
.equ    LedRecTargetVal  = 1
.equ    LedRecFadeSpeedL = 2        ; 8.8 fixed point value, byte behind point
.equ    LedRecFadeSpeedH = 3        ; byte before point
.equ    LedRecCycleCount = 4
.equ    LedRecFlags      = 5
.equ    LedRecSize       = 6

; LedRecFlags
.equ LedFlagTargetReached = 0

; serial port
.equ CLOCK = 16000000
.equ BAUD = 19200
.equ UBRRVAL = CLOCK/(BAUD*16)-1

; serial fifos
.equ URSize = 16        ; uart receive fifo size
.equ UTSize = 16        ; uart transmit fifo size

; i2c fifos
.equ IRSize = 8         ; i2c receive fifo size
.equ ITSize = 16        ; i2c transmit fifo size

; }}}

;   fifos {{{
; =============================================================================
.dseg
        ;   fifo: UReceiveFifo
        ; ==========================
        UReceiveFifo:
            .byte   FifoControlSize + URSize

        ;   fifo: UTransmitFifo
        ; ==========================
        UTransmitFifo:
            .byte   FifoControlSize + UTSize

        ;   fifo: IReceiveFifo
        ; ==========================
        IReceiveFifo:
            .byte   FifoControlSize + IRSize

        ;   fifo: ITransmitFifo
        ; ==========================
        ITransmitFifo:
            .byte   FifoControlSize + ITSize

; }}}

.cseg

;   register definitions {{{
; =============================================================================
.def    PWMBurstOutputTableL    =   r0  ; could become a problem - r0 and r1 are used by mul
.def    PWMBurstOutputTableH    =   r1  ; but we aren't using mul ATM, so it's ok...
.def    TimerTableL             =   r2
.def    TimerTableH             =   r3
.def    IReceiveFifoWPointerL   =   r4  ; i2c receive fifo temporary write pointer
.def    IReceiveFifoWPointerH   =   r5
.def    IReceiveFifoCounter     =   r6
.def    l1      =   r8      ; low working registers
.def    l2      =   r9
.def    PWMNextOutputValue  =   r10  ; pwm output value for next pulse (non-burst mode)
.def    w1      =   r16     ; working registers
.def    w2      =   r17
.def    w3      =   r18
.def    w4      =   r19
.def    w5      =   r20
.def    Pulse   =   r21     ; pwm pulse counter
.def    Flags   =   r22     ; global flags
.def    t1w1    =  r24      ; timer1 interrupt handler working registers
.def    t1w2    =  r25

; registers Y and Z may be used freely (although Z is used in the timer1 interrupt)

; }}}

;   interrupt table {{{
; =============================================================================
.ORG 0x0                            ; reset vector
        rjmp reset

.ORG OC1Aaddr                       ; timer1 CompareA interrupt handler
        rjmp OC1AHandler
.ORG URXCaddr                       ; USART receive complete interrupt handler
        rjmp URXCHandler
.ORG UDREaddr                       ; USART data register empty handler
        rjmp UDREHandler
.ORG TWIaddr                        ; TWI data handler
        rjmp TWIHandler
.ORG 0x13                           ; program start

; }}}

; We need the register definitions before including this file
.include "script_handler.inc"
.include "remote_exec_handler.inc"


;   interrupt handler
; =============================================================================

;   function: reset interrupt handler {{{
; =============================================================================
reset:

        ;  init stack
        ; ========================================
        ldi w1, low(RAMEND)
        out SPL, w1
        ldi w1, high(RAMEND)
        out SPH, w1

        ;  init outputs and set default output
        ; ========================================
        ldi w1, (1<<PB0) | (1<<PB1) | (1<<PB2)   ; configure PB0-PB2 as output
        out DDRB, w1
        ldi w1, 0b111               ; all channels high, leds off
        out PORTB, w1

        ;  init X and Y pointers
        ; ========================================
        ldi XH, high(PWMBurstOutputTable) ; load address of the output table
        ldi XL, low(PWMBurstOutputTable)
        movw PWMBurstOutputTableH:PWMBurstOutputTableL, XH:XL ; copy address into registers

        ldi YH, HIGH(LedStatusTable)   ; load LedStatusTable address into Y and registers (for use with ld)
        ldi YL, LOW(LedStatusTable)

        ;  init LedStatusTable
        ; ========================================
        ldi w1, PWMChannels         ; load channel count
    resetInitLST:
        ldi w2, 0                   ; current brightness
        st Y+, w2
        ldi w2, 0                   ; target brightness
        st Y+, w2
        ldi w2, 64                  ; LedRecFadeSpeedL
        st Y+, w2
        ldi w2, 0                   ; LedRecFadeSpeedH
        st Y+, w2
        ldi w2, 0                   ; cycle counter
        st Y+, w2
        ldi w2, 0                   ; flags
        st Y+, w2
        dec w1
        brne resetInitLST           ; loop until everything is initialized

        ;  init Z pointer
        ; ========================================
        ldi ZH, high(PWMTimerTable<<1) ; load address of the timeslot table
        ldi ZL, low(PWMTimerTable<<1)
        movw TimerTableH:TimerTableL, ZH:ZL

        ;  init flags and pulse counter
        ; ========================================
        ldi Flags, (1<<PWMPulse0)   ; first is pulse0

        ldi Pulse, 16               ; pulse counter is 16, because this counter is checked for the
                                    ; first time in pulse 16,

        ;  init timer1
        ; ========================================
        ldi w1, high(0xffff)         ; initial delay before pwm loop starts
        out OCR1AH, w1
        ldi w1, low(0xffff)
        out OCR1AL, w1

        ldi w1, (1<<CS10)|(1<<WGM12) ; no prescaler, clear timer1 on overflow
        out TCCR1B, w1
        clr w1                      ; nothing else
        out TCCR1A, w1

        ;  init uart
        ; ========================================
        ldi w1, high(UBRRVAL)
        out UBRRH, w1
        ldi w1, low(UBRRVAL)
        out UBRRL, w1
        ldi w1, (1<<URSEL)|(3<<UCSZ0) ; frame format: 8N1
        out UCSRC, w1

        ;  init i2c
        ; ========================================
        ldi w1, high(EEPROM_I2CADDR)    ; load i2c address in eeprom into EEAR[LH]
        out EEARH, w1
        ldi w1, low(EEPROM_I2CADDR)
        out EEARL, w1

        sbi EECR, EERE              ; start read

        in w1, EEDR                 ; copy read data

        cpi w1, 0xff                ; check for invalid i2c address
        brne i2c_continue_init      ; (when eeprom is not programmed)

        ldi w1, 0b110               ; toggle red leds on
        out PORTB, w1

    i2c_invalid_address:            ; refuse to do anything else
        rjmp i2c_invalid_address

    i2c_continue_init:
        out TWAR, w1                ; set i2c address
        ldi w1, (1<<TWEN)|(1<<TWINT)|(1<<TWIE)|(1<<TWEA)
        out TWCR, w1

        ;  init fifos
        ; ========================================
        MACRO_FIFO_INIT UReceiveFifo
        MACRO_FIFO_INIT UTransmitFifo
        MACRO_FIFO_INIT IReceiveFifo
        MACRO_FIFO_INIT ITransmitFifo


        ; init ScriptStatusTable
        ; ========================================
        ; TODO: script initialization
        ldi w1, ScriptHandlerFlash

        ;ldi w1, ScriptHandlerDisabled ; FIXME
        sts ScriptInfoTable+ScrRecHandler, w1
        sts ScriptInfoTable+ScrRecSize+ScrRecHandler, w1
        sts ScriptInfoTable+2*ScrRecSize+ScrRecHandler, w1

        ldi w1, low(static_script_r*2)
        sts ScriptInfoTable+ScrRecPos, w1
        ldi w1, high(static_script_r*2)
        sts ScriptInfoTable+ScrRecPos+1, w1

        ldi w1, low(static_script_g*2)
        sts ScriptInfoTable+ScrRecSize+ScrRecPos, w1
        ldi w1, high(static_script_g*2)
        sts ScriptInfoTable+ScrRecSize+ScrRecPos+1, w1

        ldi w1, low(static_script_b*2)
        sts ScriptInfoTable+2*ScrRecSize+ScrRecPos, w1
        ldi w1, high(static_script_b*2)
        sts ScriptInfoTable+2*ScrRecSize+ScrRecPos+1, w1

        ldi w1, 0
        sts ScriptInfoTable+ScrRecStackOffset, w1
        sts ScriptInfoTable+ScrRecSize+ScrRecStackOffset, w1
        sts ScriptInfoTable+2*ScrRecSize+ScrRecStackOffset, w1

        sts ScriptInfoTable+ScrRecFlags, w1
        sts ScriptInfoTable+ScrRecSize+ScrRecFlags, w1
        sts ScriptInfoTable+2*ScrRecSize+ScrRecFlags, w1

        ;  enable timer1 interrupt
        ; ========================================
        ldi w1, (1<<OCIE1A)         ; enable timer1 compare interrupt
        out TIMSK, w1

        ;  enable uart interrupt
        ; ========================================
        ldi w1, (1<<RXCIE)|(1<<RXEN)|(1<<TXEN); receiver interrupt complete enable, receiver enable,
                                        ; transmitter enable
        out UCSRB, w1

        ;  init led registers and pwm burst table
        ; ========================================
        rcall update_led_brightness  ; update led brightness registers
        rcall update_pwm_burst_table ; fill pwm burst output table

        ;  enable interrupts
        ; ========================================
        sei

        ;  output system bootup message to uart
        ; ========================================
        rcall output_string
        .db "B", 13, 10, 0

        ;  initiate script execution
        ; ========================================
        sbr Flags, (1<<ScriptEnabled)

        ;  start main procedure
        ; ========================================
        rjmp main

; }}}

;   function: timer1 interrupt handler {{{
;
; =============================================================================

    ;   macro: macro_output_next_value {{{
    ; =============================================================================
    .macro macro_output_next_value      ; load next output value from PWMBurstOutputTable and write to output    3
            ld t1w1, X+
            out PORTB, t1w1
    .endmacro

    ; }}}

    ;   macro: macro_delay {{{
    ; =============================================================================
    .macro macro_delay                  ; do nothing for 3*n cycles
            ldi t1w1, @0                ; load given value                              1
    macro_delay_loop:
            dec t1w1                    ; decrement given value                         1
            brne macro_delay_loop       ; branch if nonzero                             1 if false, 2 if true
    .endmacro

    ; }}}

OC1AHandler:                        ;                                               8
        in t1w1, SREG               ; save the status register                      1
        push t1w1                   ;                                               2
        push XL
        push XH

        sbrs Flags, PWMPulse0       ; skip next if pulse0 flag is set               1 if false, 2 if true
        rjmp PWMCheckPulseCounter   ;                                               2

                                    ; first pulse, X has been initialized
                                    ; timing values:
                                    ;.dw 2, 7, 12, 17, 22, 27, 32, 37
                                    ;.dw 42, 47, 52, 57, 62, 66, 71, 76

        ld t1w1, X+                 ; preload pulse0 value                          2
        ld t1w2, X+                 ; preload pulse1 value                          2
        out PORTB, t1w1             ; output first value                            1
        nop                         ; wait 1 cycles                                 1
        out PORTB, t1w2             ; output second value, and wait 7 cycles        1
        macro_delay 1               ; burn 3 cycles
        nop                         ; and another one, makes 4
        macro_output_next_value     ; load next value, costs 3 cycles, makes 7 cycles :), afterwards wait 12 cycles
        macro_delay 3               ; burn 9 cycles
        macro_output_next_value     ; load next value, costs 3 cycles, makes 12 cycles :), afterwards wait 17 cycles
        macro_delay 4               ; burn 12 cycles
        nop                         ; and another 2
        nop
        macro_output_next_value     ; load next value, costs 3 cycles, makes 17 cycles :), afterwards wait 22 cycles
        macro_delay 6               ; burn 18 cycles
        nop                         ; and another one
        macro_output_next_value     ; load next value, costs 3 cycles, makes 22 cycles :), afterwards wait 27 cycles
        macro_delay 8               ; burn 24 cycles
        macro_output_next_value     ; load next value, costs 3 cycles, makes 27 cycles :), afterwards wait 32 cycles
        macro_delay 9               ; burn 29 cycles
        nop
        nop
        macro_output_next_value     ; load next value, costs 3 cycles, makes 32 cycles :), afterwards wait 37 cycles
        macro_delay 11              ; burn 34
        nop
        macro_output_next_value     ; load next value, costs 3 cycles, makes 37 cycles :), afterwards wait 42 cycles
        macro_delay 13              ; burn 39
        macro_output_next_value     ; load next value, costs 3 cycles, makes 42 cycles, afterwards wait 47 cycles
        macro_delay 14              ; burn 44
        nop
        nop
        macro_output_next_value     ; afterwards wait 52 cycles
        macro_delay 16              ; burn 49
        nop
        macro_output_next_value     ; afterwards wait 57 cycles
        macro_delay 18              ; burn 54
        macro_output_next_value     ; afterwards wait 62 cycles
        macro_delay 19              ; burn 59
        nop
        nop
        macro_output_next_value     ; afterwards wait 66 cycles
        macro_delay 21
        macro_output_next_value     ; afterwards wait 71 cycles
        macro_delay 22
        nop
        nop
        macro_output_next_value     ; afterwards set the timer for the first interrupt after the burst

        push ZL                     ; store Z
        push ZH
        ldi ZH, high(PWMTimerTable<<1) ; load PWMTimerTable into Z (for use with lpm)
        ldi ZL, low(PWMTimerTable<<1)
        lpm t1w1, Z+                ; load next interrupt timeslot (cumulated burst interrupt slots) 1
        lpm t1w2, Z+                ;                                               1
        out OCR1AH, t1w2            ; set timer TOP value for next interrupt        1
        out OCR1AL, t1w1            ;                                               1
        movw TimerTableH:TimerTableL, ZH:ZL ; save current Z pointer into registers
        pop ZH                      ; restore Z
        pop ZL

        cbr Flags, (1<<PWMPulse0)   ; clear pulse0 flag                             1

        cbi UCSRB, RXCIE            ; disable uart receive complete interrupt       1
        ;in t1w1, TWCR               ; disable TWI interrupt
        ;andi t1w1, !(1<<TWIE)
        ;out TWCR, t1w1

        rjmp PWMCalculateNextValue  ;                                               2

PWMCheckPulseCounter:
        nop                         ; wait a few cycles to push the interval between timer1 compare match and pulse0 output
        nop                         ; to the same interval which is used in burst mode pulses...
        nop

        out PORTB, PWMNextOutputValue ; output value                                1

        push ZL                     ; store Z
        push ZH
        movw ZH:ZL, TimerTableH:TimerTableL ; load current timer table pointer
        lpm t1w1, Z+                ; load next interrupt timeslot                  3
        lpm t1w2, Z+                ;                                               3
        out OCR1AH, t1w2            ; set timer TOP value for next interrupt        1
        out OCR1AL, t1w1            ;                                               1
        movw TimerTableH:TimerTableL, ZH:ZL ; save current Z pointer into registers
        pop ZH                      ; restore Z
        pop ZL

        inc Pulse                   ; increment pulse counter                       1

        cpi Pulse, 42               ; check for pulse 42                            1
        brne check_pulse_255        ; if not pulse 42, jump away                    1 if false, 2 if true

        sbr Flags, (1<<PWMPulseRelax)   ; time critical part of pwm is done now - other
                                    ; interrupts are allowed now

check_pulse_255:
        cpi Pulse, 255              ; check if this is the last pulse               1
        brne PWMCalculateNextValue  ; jump away if not the last pulse               1 if false, 2 if true

                                    ; otherwise reset everything
        ldi t1w1, 16                ; pulse counter is set to 16                    1
        mov Pulse, t1w1             ;                                               1

        movw XH:XL, PWMBurstOutputTableH:PWMBurstOutputTableL ; init burst output table pointer     1

        rcall update_led_brightness ;                                                              19
        rcall update_pwm_burst_table ;                                                            272

        sbr Flags, (1<<PWMPulse0)|(1<<PWMLastPulse) ; set pulse0 flag, set last pulse flag                          1
        rjmp PWMCleanup

PWMCalculateNextValue:
        ldi t1w1, (1<<PWMChannel0)|(1<<PWMChannel1)|(1<<PWMChannel2)  ; initial output (inverted)   1
        lds t1w2, PWMCurrentBrightnessValues+0    ; load channel0 desired brightness value          2
        cp Pulse, t1w2              ; channel 0 brightness reached, Pulse >= ch0?                   1
        brsh check_channel1         ; if true, jump away                                            1 if false, 2 if true
        cbr t1w1, (1<<PWMChannel0)  ; if false, clear bit for channel0                              1
check_channel1:
        lds t1w2, PWMCurrentBrightnessValues+1    ; load channel1 desired brightness value          2
        cp Pulse, t1w2              ; channel 1 brightness reached, Pulse >= ch1                    1
        brsh check_channel2         ; if true, jump away                                            1 if false, 2 if true
        cbr t1w1, (1<<PWMChannel1)  ; if false, clear bit for channel1                              1
check_channel2:
        lds t1w2, PWMCurrentBrightnessValues+2    ; load channel2 desired brightness value          2
        cp Pulse, t1w2              ; channel 2 brightness reached, Pulse >= ch2                    1
        brsh store_next_value       ; if true, jump away                                            1 if false, 2 if true
        cbr t1w1, (1<<PWMChannel2)  ; if false, clear bit for channel2                              1
store_next_value:
        mov PWMNextOutputValue, t1w1

PWMCleanup:
        pop XH
        pop XL
        pop t1w1                    ; restore the status register                   2
        out SREG, t1w1              ;                                               1
        reti                        ;                                               4


update_led_brightness:              ; {{{ duration: 19 cycles                                       3
        lds t1w1, LedStatusTable+LedRecBrightness+LedRecSize*0 ; load brightness value                2
        sts PWMCurrentBrightnessValues+0, t1w1 ; save brightness value                                2
        lds t1w1, LedStatusTable+LedRecBrightness+LedRecSize*1 ; load brightness value                2
        sts PWMCurrentBrightnessValues+1, t1w1 ; save brightness value                                2
        lds t1w1, LedStatusTable+LedRecBrightness+LedRecSize*2 ; load brightness value                2
        sts PWMCurrentBrightnessValues+2, t1w1 ; save brightness value                                2

        ret                         ;                                                               4
        ; }}}


update_pwm_burst_table:             ; {{{ duration: 276 cycles                                      3
        push YL                     ; save Y                                                        2
        push YH                     ;                                                               2
        push w3                     ; save w3                                                       2
        movw YH:YL, PWMBurstOutputTableH:PWMBurstOutputTableL ; load address of PWMBurstOutputTable into Y  1
        ldi t1w1, 0                 ; init pulse counter                                            1
update_pwm_burst_table_loop:
        ldi t1w2, 0xff              ; initial output value (inverted)                               1
update_channel0:
        lds w3, PWMCurrentBrightnessValues+0    ; load channel0 desired brightness value            2
        cp t1w1, w3                 ; channel 0 brightness reached, w1 >= ch0?                      1
        brsh update_channel1        ; if true, jump away                                            1 if false, 2 if true
        cbr t1w2, (1<<PWMChannel0)  ; if false, clear bit for channel0                              1
update_channel1:
        lds w3, PWMCurrentBrightnessValues+1    ; load channel1 desired brightness value            2
        cp t1w1, w3                 ; channel 1 brightness reached, w1 >= ch1?                      1
        brsh update_channel2        ; if true, jump away                                            1 if false, 2 if true
        cbr t1w2, (1<<PWMChannel1)  ; if false, clear bit for channel1                              1
update_channel2:
        lds w3, PWMCurrentBrightnessValues+2    ; load channel2 desired brightness value            2
        cp t1w1, w3                 ; channel 2 brightness reached, w1 >= ch2?                      1
        brsh update_store_output    ; if true, jump away                                            1 if false, 2 if true
        cbr t1w2, (1<<PWMChannel2)  ; if false, clear bit for channel2                              1
update_store_output:
        st Y+, t1w2                 ; store inverted output value                                   2
        inc t1w1                    ;                                                               1
        cpi t1w1, 16                ;                                                               1
        brne update_pwm_burst_table_loop ;                                                          1 if false, 2 if true
        pop w3                      ; restore w3                                                    2
        pop YH                      ; restore Y                                                     2
        pop YL                      ;                                                               2
        ret                         ;                                                               4
        ; }}}

; }}}

;   function: uart receive complete interrupt handler {{{
; =============================================================================
URXCHandler:                        ;                                           8
        push w1                     ; save w1                                   2
        in w1, SREG                 ; save the status register                  1
        push w1                     ;                                           2

        MACRO_FIFO_LOAD_COUNTER UReceiveFifo, w1 ; load counter
        cpi w1, URSize              ; compare counter and fifo size             1
        breq URXDiscardData         ; if counter == fifosize, just discard the byte received  1 false/2 true

        push w2                     ; save some more registers                  2
        push YL                     ; save Y                                    2
        push YH                     ;                                           2

        in w1, UDR                  ; read data byte                            1

                                    ; store data into receive fifo
        MACRO_FIFO_STORE UReceiveFifo, w1, w2, w1, URSize

URXCleanup:
        pop YH                      ; restore Y                                 2
        pop YL                      ;                                           2
        pop w2                      ;                                           2
        pop w1                      ; restore the status register               2
        out SREG, w1                ;                                           1
        pop w1                      ; restore w1                                2
        reti                        ;                                           4

URXDiscardData:
        in w1, UDR                  ; clear interrupt by reading the received byte
        pop w1                      ; restore the status register
        out SREG, w1
        pop w1                      ; restore w1
        reti

; }}}

;   function: uart data register empty handler {{{
; =============================================================================
UDREHandler:
        push w1                     ; save w1
        in w1, SREG                 ; save the status register
        push w1

        MACRO_FIFO_LOAD_COUNTER UTransmitFifo, w1   ; load uart transmit fifo counter into w1
        cpi w1, 1
        brsh UDRETransmitByte       ; if w1 >= 1, jump away
                                    ; else:
        in w1, UCSRB                ; disable data register empty interrupt
        cbr w1, (1<<UDRIE)
        out UCSRB, w1

        pop w1                      ; restore the status register
        out SREG, w1
        pop w1                      ; restore w1
        reti

UDRETransmitByte:
        push w2                     ; save registers
        push w3
        push YL                     ; save Y
        push YH

        MACRO_FIFO_READ UTransmitFifo, w1, w2, w3, UTSize
        out UDR, w1                 ; write data to uart data register

        pop YH                      ; restore Y
        pop YL
        pop w3                      ; restore registers
        pop w2
        pop w1                      ; restore the status register
        out SREG, w1
        pop w1                      ; restore w1
        reti
; }}}

;   function: TWI interrupt handler {{{
;
; =============================================================================
TWIHandler:
        push w1
        in w1, SREG
        push w1
        push w2
        push w3

        in w1, TWSR                     ; load twi status
        andi w1, 0b11111000             ; mask out prescaler bits

        cpi w1, 0x60                    ; sla+w+ack received
        breq normal_sla_w
        cpi w1, 0x68                    ; lost abitration in sla+r/w and own sla+w+ack received
        breq normal_sla_w
        cpi w1, 0x70                    ; gca+w+ack received
        breq normal_sla_w
        cpi w1, 0x78                    ; lost abitration in sla+r/w and gca+w+ack received
        breq normal_sla_w
        cpi w1, 0x80                    ; data received as sla+w, ack
        breq normal_sla_w_data_ack
        cpi w1, 0x88                    ; data received as sla+w, nack
        breq normal_sla_stop_bridge
        cpi w1, 0x90                    ; data received as gca+w, ack
        breq normal_sla_w_data_ack
        cpi w1, 0x98                    ; data received as gca+w, nack
        breq normal_sla_stop_bridge
        cpi w1, 0xA0
        breq normal_sla_stop_bridge

        .ifdef SERIALDEBUG
            ldi w1, 'X'
            rcall write_uart_transmit_fifo
        .endif

        rjmp normal_end

normal_sla_w: ; {{{
        push YL                             ; save Y
        push YH

        .ifdef SERIALDEBUG
            ldi w1, 'S'
            rcall write_uart_transmit_fifo
        .endif

        lds w1, IReceiveFifo + FifoWPointerL ; load ireceivefifo write pointer
        lds w2, IReceiveFifo + FifoWPointerH
        movw IReceiveFifoWPointerH:IReceiveFifoWPointerL, w2:w1 ; save ireceivefifo write pointer

        ldi w1, 0
        MACRO_FIFO_STORE IReceiveFifo, w1, w2, w1, IRSize  ; store nullbyte

        mov IReceiveFifoCounter, w1         ; reset data counter

        pop YH                              ; restore Y
        pop YL

        rjmp normal_end
; }}}

normal_sla_stop_bridge:
        rjmp normal_sla_stop

normal_sla_w_data_ack: ; {{{
        push YL                             ; save Y
        push YH

        in w1, TWDR
        .ifdef SERIALDEBUG
                mov w3, w1
        .endif

        MACRO_FIFO_STORE IReceiveFifo, w1, w2, w1, IRSize  ; store data byte

        mov w1, IReceiveFifoCounter     ; increment counter
        inc w1
        mov IReceiveFifoCounter, w1

        .ifdef SERIALDEBUG
            ldi w1, 'D'
            rcall write_uart_transmit_fifo
            mov w1, w3
            rcall write_uart_transmit_fifo
        .endif

        pop YH                              ; restore Y
        pop YL

        rjmp normal_end
; }}}

normal_sla_stop: ; {{{
        push YL                             ; save Y
        push YH

        .ifdef SERIALDEBUG
            ldi w1, 's'
            rcall write_uart_transmit_fifo
        .endif

        mov w1, IReceiveFifoCounter     ; load counter

        cpi w1, 0                       ; check if we received something
        breq normal_sla_stop_reset_fifo

        .ifdef SERIALDEBUG
            push w1
            ldi w1, 'R'
            rcall write_uart_transmit_fifo
            pop w1
            push w1
            rcall write_uart_transmit_fifo
            pop w1
        .endif

        movw YH:YL, IReceiveFifoWPointerH:IReceiveFifoWPointerL ; else save counter into byte already on the fifo
        std Y+0, w1

        pop YH                              ; restore Y
        pop YL

        rjmp normal_end

    normal_sla_stop_reset_fifo:
        MACRO_FIFO_POP IReceiveFifo, w1, w2, w3, IRSize ; pop last written (null)byte

        pop YH                              ; restore Y
        pop YL

        .ifdef SERIALDEBUG
            push w1
            ldi w1, 'r'
            rcall write_uart_transmit_fifo
            pop w1
        .endif

        rjmp normal_end

; }}}

normal_end: ; {{{
        in w1, TWCR
        ori w1, (1<<TWINT)|(1<<TWEA)
        out TWCR, w1

        pop w3
        pop w2
        pop w1
        out SREG, w1
        pop w1
        reti
; }}}

; }}}


;   other functions
; =============================================================================

;   function: main {{{
; =============================================================================
main:
        sbrc Flags, PWMLastPulse    ;  check for pwm pulse end
        rjmp main_last_pulse

        sbrc Flags, PWMPulseRelax
        rjmp main_pulse_relax

        rcall check_serial_input    ; check for serial input via uart

        rjmp main

main_last_pulse:
        ;  update status table
        ; =================================================
        rcall update_status_table

        cbr Flags, (1<<PWMLastPulse)    ; clear burst-end flag
        rjmp main

main_pulse_relax:
        sbi UCSRB, RXCIE            ; enable uart receive complete interrupt
        ;in t1w1, TWCR               ; enable TWI interrupt
        ;ori t1w1, (1<<TWIE)
        ;out TWCR, t1w1

        rcall execute_remote_commands   ; remote commands

        sbrc Flags, ScriptEnabled       ; scripts
        rcall execute_scripts

        cbr Flags, (1<<PWMPulseRelax)   ; clear flag
        rjmp main

; }}}


;   function: output_string {{{
; =============================================================================
output_string:
        pop ZH              ; load string address from the stack
        pop ZL
        lsl ZL              ; shift left Z (using carry)
        rol ZH
output_next_char:
        lpm w1, Z+          ; load next byte into w1 and increment Z
        cpi w1, 0           ; if it's not the end of the string
        breq output_string_end ; jump away

        rcall write_uart_transmit_fifo ; transmit char
        rjmp output_next_char ; loop

output_string_end:
        adiw ZL, 1          ; increment Z to point to the next instruction after the string
        lsr ZH              ; shift right Z (using carry)
        ror ZL
        push ZL             ; push Z back onto the stack
        push ZH
        ret

; }}}

;   function: update_status_table {{{
; =============================================================================
update_status_table:
        ldi YH, high(LedStatusTable)                     ; load LedStatusTable address into Y
        ldi YL, low(LedStatusTable)
        ldi w1, 3
        mov l1, w1

us_update_channels:                                      ; update all three channels
        ; loop body: update current channel {{{
        ldd w1, Y+LedRecBrightness                  ; load channel brightness
        ldd w2, Y+LedRecTargetVal                   ; load channel target value
        cp w2, w1
        breq us_cleanup                             ; if equal, jump to cleanup, nothing to do
        ldd w3, Y+LedRecCycleCount                  ; load cycle counter

        ldd w4, Y+LedRecFadeSpeedL                  ; load fade speed, byte below 1
        ldd w5, Y+LedRecFadeSpeedH                  ; load fade speed, byte above 1
        add w3, w4                                  ; add low byte
        clr w4                                      ; clear w4
        adc w4, w5                                  ; add high byte, now w4 is brightstep.

        std Y+LedRecCycleCount, w3                  ; store low byte in cycle count for next round

        cp w2, w1
        brlo us_decrement                           ; if target val < brightness, jump to decrement part

        add w1, w4                                  ; increase brightness
        brcs us_target_missed                       ; if an overflow happened, just set the target value

        cp w2, w1                                   ; compare brightness after increase and target brigtness
        brsh us_store                               ; if current <= target, store brightness
        rjmp us_target_missed                       ; else we added too much, just set the target value...

us_decrement:
        sub w1, w4                                  ; decrease brightness
        brcs us_target_missed                       ; if an underflow happened, just set the target value

        cp w1, w2                                   ; compare current and target value
        brsh us_store                               ; if current >= target, store current brightness
                                                    ; else we missed the target

us_target_missed:
        mov w1, w2                                  ; set the target value

us_store:
        cp w1, w2                                   ; compare current brightness and target brightness
        brne us_store2                              ; if not equal, just store the new brightness value
        ldd w4, Y+LedRecFlags                       ; load led flags
        sbr w4, (1<<LedFlagTargetReached)           ; set the target reached flag
        std Y+LedRecFlags, w4                       ; store the new flags
us_store2:
        std Y+LedRecBrightness, w1                  ; store brightness

us_cleanup:
        ; }}}
        adiw YH:YL, LedRecSize
        dec l1
        brne us_update_channels
        ret

; }}}

;   function: check_serial_input {{{
; =============================================================================
check_serial_input:
        MACRO_FIFO_LOAD_COUNTER UReceiveFifo, w1   ; load uart receive fifo counter into w1
        cpi w1, 0
        breq check_ret1 ; if not, return

        MACRO_FIFO_READ UReceiveFifo, w1, w2, w3, URSize   ; else load received byte into w1

        ldi YH, high(LedStatusTable)                     ; load LedStatusTable address into Y
        ldi YL, low(LedStatusTable)

        rjmp check

check_ret1:
        ret

check:
        ldi w2, 5

        cpi w1, '1'
        brne check_4

        ldd w1, Y+LedRecTargetVal+0*LedRecSize
        sub w1, w2
        std Y+LedRecTargetVal+0*LedRecSize, w1
        rjmp check_end

check_4:
        cpi w1, '4'
        brne check_2

        ldd w1, Y+LedRecTargetVal+0*LedRecSize
        add w1, w2
        std Y+LedRecTargetVal+0*LedRecSize, w1
        rjmp check_end

check_2:
        cpi w1, '2'
        brne check_5

        ldd w1, Y+LedRecTargetVal+1*LedRecSize
        sub w1, w2
        std Y+LedRecTargetVal+1*LedRecSize, w1
        rjmp check_end

check_5:
        cpi w1, '5'
        brne check_3

        ldd w1, Y+LedRecTargetVal+1*LedRecSize
        add w1, w2
        std Y+LedRecTargetVal+1*LedRecSize, w1
        rjmp check_end

check_3:
        cpi w1, '3'
        brne check_6

        ldd w1, Y+LedRecTargetVal+2*LedRecSize
        sub w1, w2
        std Y+LedRecTargetVal+2*LedRecSize, w1
        rjmp check_end

check_6:
        cpi w1, '6'
        brne check_0

        ldd w1, Y+LedRecTargetVal+2*LedRecSize
        add w1, w2
        std Y+LedRecTargetVal+2*LedRecSize, w1
        rjmp check_end

check_0:
        cpi w1, '0'
        brne check_enter

        ldi w1, 0
        std Y+LedRecTargetVal+0*LedRecSize, w1
        std Y+LedRecTargetVal+1*LedRecSize, w1
        std Y+LedRecTargetVal+2*LedRecSize, w1
        rjmp check_end

check_enter:
        cpi w1, 13
        brne check_plus

        ldd w1, Y+LedRecBrightness+0*LedRecSize
        std Y+LedRecTargetVal+0*LedRecSize, w1
        ldd w1, Y+LedRecBrightness+1*LedRecSize
        std Y+LedRecTargetVal+1*LedRecSize, w1
        ldd w1, Y+LedRecBrightness+2*LedRecSize
        std Y+LedRecTargetVal+2*LedRecSize, w1
        rjmp check_end

check_plus:
        cpi w1, '+'
        brne check_minus

        ldd w1, Y+LedRecBrightness+0*LedRecSize
        inc w1
        std Y+LedRecTargetVal+0*LedRecSize, w1
        ldd w1, Y+LedRecBrightness+1*LedRecSize
        inc w1
        std Y+LedRecTargetVal+1*LedRecSize, w1
        ldd w1, Y+LedRecBrightness+2*LedRecSize
        inc w1
        std Y+LedRecTargetVal+2*LedRecSize, w1
        rjmp check_end

check_minus:
        cpi w1, '-'
        brne check_pgm

        ldd w1, Y+LedRecBrightness+0*LedRecSize
        dec w1
        std Y+LedRecTargetVal+0*LedRecSize, w1
        ldd w1, Y+LedRecBrightness+1*LedRecSize
        dec w1
        std Y+LedRecTargetVal+1*LedRecSize, w1
        ldd w1, Y+LedRecBrightness+2*LedRecSize
        dec w1
        std Y+LedRecTargetVal+2*LedRecSize, w1
        rjmp check_end

check_pgm:
        cpi w1, 'x'
        brne check_script
        rjmp 0xc00

check_script:
        cpi w1, 's'
        brne check_script_r

        ldi w2, (1<<ScriptEnabled)
        eor Flags, w2
        rjmp check_end

check_script_r:
        cpi w1, 'r'
        brne check_script_g

        lds w1, ScriptInfoTable+ScrRecHandler
        ldi w2, 0b10000000
        eor w1, w2
        sts ScriptInfoTable+ScrRecHandler, w1

        rjmp check_end

check_script_g:
        cpi w1, 'g'
        brne check_script_b

        lds w1, ScriptInfoTable+ScrRecSize+ScrRecHandler
        ldi w2, 0b10000000
        eor w1, w2
        sts ScriptInfoTable+ScrRecSize+ScrRecHandler, w1

        rjmp check_end

check_script_b:
        cpi w1, 'b'
        brne check_end

        lds w1, ScriptInfoTable+2*ScrRecSize+ScrRecHandler
        ldi w2, 0b10000000
        eor w1, w2
        sts ScriptInfoTable+2*ScrRecSize+ScrRecHandler, w1

        rjmp check_end


check_end:
        ret
; }}}

;   function: write_uart_transmit_fifo {{{
; =============================================================================
; data byte is loaded into w1
write_uart_transmit_fifo:
        MACRO_FIFO_LOAD_COUNTER UTransmitFifo, w2   ; load uart transmit fifo counter into w2
        cpi w2, UTSize ; compare counter and fifo size
        breq write_uart_discard_data ; if counter == fifosize, just discard the byte received

        movw l2:l1, YH:YL           ; save Y

        MACRO_FIFO_STORE UTransmitFifo, w1, w2, w1, UTSize
        sbi UCSRB, UDRIE            ; enable data register empty interrupt
        movw YH:YL, l2:l1           ; restore Y

write_uart_discard_data:

        ret
; }}}


;   tables
; =============================================================================

;   data segment(ram), dynamic tables
; =============================================================================
.dseg

;   table: PWMCurrentBrightnessValues {{{
; =============================================================================
PWMCurrentBrightnessValues:
        .byte PWMChannels

; }}}

;   table: PWMBurstOutputTable {{{
; =============================================================================
; The burst table contains a bitmask for the LEDs for each pwm cycle in burst
; mode

PWMBurstOutputTable:
        .byte   PWMBurstLevels

; }}}

;   table: LedStatusTable {{{
; =============================================================================
; The status table contains a record for each LED.
; This record consists of
;   .byte   : (LedRedBrightness) Current brightness value
;   .byte   : (LedRecTargetVal)  Target brightness value (0-255)
;   .byte   : (LedRecFadeSpeedL) 8.8 fixed point value
;   .byte   : (LedRecFadeSpeedH)
;   .byte   : (LedRecCycleCount) Cycle counter
;   .byte   : (LedRecFlags)      Flags: TargetReached, ...

LedStatusTable:
        .byte   PWMChannels * LedRecSize

; }}}

;   program segment(flash), static tables
; =============================================================================
.cseg

;   table: PWMTimerTable {{{
; =============================================================================
; timer1 TOP values
PWMTimerTable:
        ; burst timeslots:
        ;.dw 2, 7, 12, 17, 22, 27, 32, 37
        ;.dw 42, 47, 52, 57, 62, 66, 71, 76
        ; normal timeslots, first is the cumulated burst timeslot
        .dw 629, 81, 86, 91, 96, 101, 106, 111
        .dw 116, 121, 125, 130, 135, 140, 145, 150
        .dw 155, 160, 165, 170, 175, 180, 185, 189
        .dw 194, 199, 204, 209, 214, 219, 224, 229
        .dw 234, 239, 244, 249, 253, 258, 263, 268
        .dw 273, 278, 283, 288, 293, 298, 303, 308
        .dw 312, 317, 322, 327, 332, 337, 342, 347
        .dw 352, 357, 362, 367, 372, 376, 381, 386
        .dw 391, 396, 401, 406, 411, 416, 421, 426
        .dw 431, 436, 440, 445, 450, 455, 460, 465
        .dw 470, 475, 480, 485, 490, 495, 500, 504
        .dw 509, 514, 519, 524, 529, 534, 539, 544
        .dw 549, 554, 559, 563, 568, 573, 578, 583
        .dw 588, 593, 598, 603, 608, 613, 618, 623
        .dw 627, 632, 637, 642, 647, 652, 657, 662
        .dw 667, 672, 677, 682, 687, 691, 696, 701
        .dw 706, 711, 716, 721, 726, 731, 736, 741
        .dw 746, 750, 755, 760, 765, 770, 775, 780
        .dw 785, 790, 795, 800, 805, 810, 814, 819
        .dw 824, 829, 834, 839, 844, 849, 854, 859
        .dw 864, 869, 874, 878, 883, 888, 893, 898
        .dw 903, 908, 913, 918, 923, 928, 933, 937
        .dw 942, 947, 952, 957, 962, 967, 972, 977
        .dw 982, 987, 992, 997, 1001, 1006, 1011, 1016
        .dw 1021, 1026, 1031, 1036, 1041, 1046, 1051, 1056
        .dw 1061, 1065, 1070, 1075, 1080, 1085, 1090, 1095
        .dw 1100, 1105, 1110, 1115, 1120, 1124, 1129, 1134
        .dw 1139, 1144, 1149, 1154, 1159, 1164, 1169, 1174
        .dw 1179, 1184, 1188, 1193, 1198, 1203, 1208, 1213
        .dw 1218, 1223, 1228, 1233, 1238, 1243, 1248, 1252

; }}}


;   static scripts
; =============================================================================

.include "script_macros.inc"
.include "static_scripts.inc"
